import {
  RPCClientCodec,
  RPCClientResponseCodec,
  RPCClientNotificationCodec,
} from '../RPCClientCodec';

import {
  RPCError,
} from '../RPCError';
import { RPCMessage } from '../../core/type';
import { CodecSelector, defaultCodecSelector } from '../../codec/codec';

export interface ClientNotification {
  method: string;
  params?: string[];
}

export interface ClientRequest {
  jsonrpc: string;
  method: string;
  params?: string[];
  id?: number;
}

export interface ClientResponse {
  jsonrpc: string;
  result?: any;
  error?: RPCError;
  id?: number;
}

export class JSONRPCClientCodec implements RPCClientCodec {
  private readonly codecSelector: CodecSelector;

  public constructor(codecSelector: CodecSelector = defaultCodecSelector) {
    this.codecSelector = codecSelector;
  }

  public request(method: string, args: any[], id?: number): RPCMessage {
    const params = convertParamsToStringArray(args);

    const req: ClientRequest = {
      jsonrpc: '2.0',
      method: method,
      params: 0 === params.length ? null : params,
      id: id,
    };

    return this.codecSelector.encode(JSON.stringify(req));
  }

  public response(message: RPCMessage): RPCClientResponseCodec {
    return new JSONRPCClientResponseCodec(this.codecSelector.decode(message));
  }
}

function convertParamsToStringArray(args: any[]): string[] | undefined {
  const values: string[] = [];

  if (undefined === args || null === args || 0 === args.length) {
    return values;
  }

  for (let indexI = 0; indexI < args.length; indexI++) {
    const arg = args[indexI];

    switch (typeof arg) {
      case 'boolean':
      case 'number': // enum
        values.push(String(arg));
        break;
      case 'string':
        values.push(arg);
        break;
      case 'object':  // array, map
        values.push(JSON.stringify(arg));
        break;
      default:
        throw new Error(`Not supported type[${typeof arg}]`);
    }
  }

  return values;
}

export class JSONRPCClientResponseCodec implements RPCClientResponseCodec {
  private res: ClientResponse;

  public constructor(json: string) {
    this.res = JSON.parse(json);
  }

  public id(): number | undefined {
    return this.res.id;
  }
  public error(): RPCError | undefined {
    return this.res.error;
  }
  public result(): any | undefined {
    return this.res.result;
  }

  public isNotification(): boolean {
    if (undefined === this.id() && undefined !== this.result()) {
      return true;
    }
    return false;
  }

  public notification(): RPCClientNotificationCodec | undefined {
    if (undefined !== this.id() || undefined === this.result()) {
      return undefined;
    }
    const noti: ClientNotification = {
      method: this.res.result.method,
      params: this.res.result.params,
    };
    return new JSONRPCClientNotificationCodec(noti);
  }
}

export class JSONRPCClientNotificationCodec implements RPCClientNotificationCodec {
  public constructor(private noti: ClientNotification) {
  }

  public method(): string {
    return this.noti.method;
  }
  public params(): any[] | undefined {
    return this.noti.params;
  }
}
